@@ -41,12 +41,12 @@ import sys
 import time
 
 import gi
-gi.require_version("Gdk", "3.0")
-gi.require_version("Gtk", "3.0")
-gi.require_version("Handy", "1")
+gi.require_version("Gdk", "4.0")
+gi.require_version("Gtk", "4.0")
+gi.require_version("Adw", "1")
 gi.require_version("PackageKitGlib", "1.0")
 from gi.repository import PackageKitGlib as packagekit
-from gi.repository import GObject, Gdk, Gtk, Gio, GLib, Handy
+from gi.repository import GObject, Gdk, Gtk, Gio, GLib, Adw
 
 from aptsources.sourceslist import Deb822SourceEntry
 
@@ -69,6 +69,7 @@ from softwareproperties.gtk.utils import (
     get_ua_service_status,
     current_distro,
     is_current_distro_lts,
+    maybe_hint_dialog_window,
 )
 
 from UbuntuDrivers import detect
@@ -167,16 +168,12 @@ def maybe_log_authentication_canceled_error(dbus_exception):
 
 
 def error(parent_window, summary, msg):
-    """ show a error dialog """
-    dialog = Gtk.MessageDialog(parent=parent_window,
-                               flags=Gtk.DialogFlags.MODAL,
-                               type=Gtk.MessageType.ERROR,
-                               buttons=Gtk.ButtonsType.OK,
-                               message_format=None)
-    dialog.set_markup("<big><b>%s</b></big>\n\n%s" % (summary, msg))
-    dialog.run()
-    dialog.destroy()
-    return False
+    """Show an error dialog."""
+    dlg = Adw.MessageDialog.new(parent_window, str(summary), str(msg))
+    dlg.add_response("ok", _("OK"))
+    dlg.set_close_response("ok")
+    dlg.connect("response", lambda d, r: d.hide())
+    dlg.present()
 
 
 class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
@@ -190,17 +187,32 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         Gtk.Window.set_default_icon_name("software-properties")
 
         # Support GNOME 42 dark mode
-        Handy.init()
-        style_manager = Handy.StyleManager.get_default()
-        style_manager.set_color_scheme(Handy.ColorScheme.PREFER_LIGHT)
+        style_manager = Adw.StyleManager.get_default()
+        style_manager.set_color_scheme(Adw.ColorScheme.PREFER_LIGHT)
 
         SimpleGtkbuilderApp.__init__(self,
                                      os.path.join(datadir, "gtkbuilder", "main.ui"),
                                      domain="software-properties")
 
+        # GTK4: Manually connect signals.
+        # Window lifecycle
+        self.window_main.connect("close-request", self.on_close_request)
+        # Main action buttons
+        self.button_close.connect("clicked", self.on_close_button)
+        self.button_revert.connect("clicked", self.on_button_revert_clicked)
+        # Other Software tab
+        self.button_add.connect("clicked", self.on_add_clicked)
+        self.button_edit.connect("clicked", self.on_edit_clicked)
+        self.button_remove.connect("clicked", self.on_remove_clicked)
+        self.treeview_sources.connect("cursor-changed", self.on_treeview_sources_cursor_changed)
+        # Authentication tab
+        self.button_auth_add.connect("clicked", self.add_key_clicked)
+        self.button_auth_remove.connect("clicked", self.remove_key_clicked)
+        # Additional Drivers tab
+        self.button_drivers_control.connect("clicked", self.on_detect_drivers_cancel_retry)
+
         if parent:
-            self.window_main.set_type_hint(Gdk.WindowTypeHint.DIALOG)
-            self.window_main.show()
+            maybe_hint_dialog_window(self.window_main)
             try:
                 self.window_main.set_transient_for(parent)
             except:
@@ -209,8 +221,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         # If externally called, reparent to external application.
         self.options = options
         if options and options.toplevel != None:
-            self.window_main.set_type_hint(Gdk.WindowTypeHint.DIALOG)
-            self.window_main.show()
+            maybe_hint_dialog_window(self.window_main)
             try:
                 toplevel = Gdk.window_foreign_new(int(options.toplevel))
             except AttributeError:
@@ -248,9 +259,6 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         # Reload dbus backend, for if there are any sources changed.
         self.backend.Reload();
 
-        # Show what we have early
-        self.window_main.show()
-
         # Did detect_drivers already run normally?
         self.successfully_detected_drivers = False
 
@@ -302,7 +310,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             self.open_file(file)
 
         # Make sure window is as small as possible initially.
-        self.window_main.resize(1, 1)
+        self.window_main.set_default_size(1, 1)
 
     def do_shutdown(self):
         if self.search_process and self.search_process.is_alive():
@@ -331,6 +339,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
 
     def init_release_upgrades(self):
         " setup the widgets that allow configuring the release upgrades "
+        self.maybe_disconnect_handler(self.combobox_release_upgrades)
         i = self.get_release_upgrades_policy()
         self.combobox_release_upgrades.set_active(i)
         self.handlers[self.combobox_release_upgrades] = \
@@ -350,6 +359,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
 
     def init_auto_update(self):
         """ Set up the widgets that allow to configure the update automation """
+        self.maybe_disconnect_handler(self.combobox_update_interval)
         self.combobox_update_interval.show()
 
         # normal updates
@@ -381,6 +391,11 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
              self.combobox_update_interval.connect("changed",
                                                    self.on_combobox_update_interval_changed)
 
+    def maybe_disconnect_handler(self, owner):
+        handler = self.handlers.get(owner)
+        if handler:
+            owner.disconnect(handler)
+
     def show_auto_update_level(self):
         """Represent the level of update automation in the user interface"""
 
@@ -397,10 +412,12 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                     self.combobox_other_updates.set_active(i)
                     break
 
+        self.maybe_disconnect_handler(self.combobox_security_updates)
         self.handlers[self.combobox_security_updates] = \
              self.combobox_security_updates.connect("changed",
                                                     self.set_sec_update_automation_level)
 
+        self.maybe_disconnect_handler(self.combobox_other_updates)
         self.handlers[self.combobox_other_updates] = \
              self.combobox_other_updates.connect("changed",
                                                  self.set_other_update_automation_level)
@@ -423,6 +440,11 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         elif level_sec == softwareproperties.UPDATE_INST_SEC:
             self.combobox_security_updates.set_active(2)  # Download and install automatically
 
+    def remove_all_box_children(self, box):
+        c = box.get_first_child()
+        while c:
+            box.remove(c)
+            c = box.get_first_child()
 
     def init_distro(self):
         """Setup the user interface elements to represent the distro"""
@@ -435,8 +457,8 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                                                                    self.on_checkbutton_source_code_toggled)
 
         # Setup the checkbuttons for the components
-        for checkbutton in self.vbox_dist_comps.get_children():
-            self.vbox_dist_comps.remove(checkbutton)
+        self.remove_all_box_children(self.vbox_dist_comps)
+
         for comp in self.distro.source_template.components:
             # TRANSLATORS: Label for the components in the Internet section
             #              first %s is the description of the component
@@ -449,12 +471,12 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             self.handlers[checkbox] = checkbox.connect("toggled",
                                                    self.on_component_toggled,
                                                    comp.name)
-            self.vbox_dist_comps.add(checkbox)
+            self.vbox_dist_comps.append(checkbox)
             checkbox.show()
 
         # Setup the checkbuttons for the child repos
-        for checkbutton in self.dev_box.get_children():
-            self.dev_box.remove(checkbutton)
+        self.remove_all_box_children(self.dev_box)
+
         for template in self.distro.source_template.children:
             # Do not show source entries in there
             if template.type == "deb-src":
@@ -467,7 +489,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                     checkbox.connect("toggled",
                                      self.on_checkbutton_child_toggled,
                                      template)
-                self.dev_box.add(checkbox)
+                self.dev_box.append(checkbox)
                 checkbox.show()
 
         status = get_ua_status()
@@ -523,6 +545,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
 
         # Setup the combo box for updates subscriptions
         combobox = self.combobox_updates_subscription
+        self.maybe_disconnect_handler(combobox)
         self.handlers[combobox] = combobox.connect(
             "changed", self.on_combobox_updates_subscription_changed)
 
@@ -530,6 +553,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         cell = Gtk.CellRendererText()
         self.combobox_server.pack_start(cell, True)
         self.combobox_server.add_attribute(cell, 'text', 0)
+        self.maybe_disconnect_handler(self.combobox_server)
         self.handlers[self.combobox_server] = self.combobox_server.connect("changed",
                                                            self.on_combobox_server_changed)
         server_store = Gtk.ListStore(GObject.TYPE_STRING,
@@ -553,17 +577,21 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         self.block_handlers()
 
         # Enable or disable the child source checkbuttons
-        for checkbox in self.dev_box.get_children():
+        checkbox = self.dev_box.get_first_child()
+        while checkbox:
             (active, inconsistent) = self.get_comp_child_state(checkbox.template)
             checkbox.set_active(active)
             checkbox.set_inconsistent(inconsistent)
+            checkbox = checkbox.get_next_sibling()
 
         # Enable or disable the component checkbuttons
-        for checkbox in self.vbox_dist_comps.get_children():
+        checkbox = self.vbox_dist_comps.get_first_child()
+        while checkbox:
             # check if the comp is enabled
             (active, inconsistent) = self.get_comp_download_state(checkbox.comp)
             checkbox.set_inconsistent(inconsistent)
             checkbox.set_active(active)
+            checkbox = checkbox.get_next_sibling()
 
         # Refresh the combo box for updates subscriptions
         subscriptions = UpdateSubscriptions()
@@ -680,7 +708,23 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         # only set if something actually changed
         try:
             if state != self.get_update_automation_level():
-                self.backend.SetUpdateAutomationLevel(state)
+                def handler():
+                    widget.set_sensitive(True)
+
+                def on_error(e):
+                    widget.set_sensitive(True)
+
+                    if isinstance(e, dbus.DBusException):
+                        if maybe_log_authentication_canceled_error(e):
+                            self.show_auto_update_level()
+                            return
+                    raise e
+
+                widget.set_sensitive(False)
+                self.backend.SetUpdateAutomationLevel(state,
+                                                      reply_handler=handler,
+                                                      error_handler=on_error)
+
         except dbus.DBusException as e:
             if maybe_log_authentication_canceled_error(e):
                 combo_handler = self.handlers[self.combobox_security_updates]
@@ -702,17 +746,23 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
 
     def on_combobox_release_upgrades_changed(self, combobox):
         """ set the release upgrades policy """
-        #print("on_combobox_release_upgrades_changed()")
         i = combobox.get_active()
-        try:
-            self.backend.SetReleaseUpgradesPolicy(i)
-        except dbus.DBusException as e:
-            if maybe_log_authentication_canceled_error(e):
-                combo_handler = self.handlers[self.combobox_release_upgrades]
-                self.combobox_release_upgrades.handler_block(combo_handler)
-                i = self.get_release_upgrades_policy()
-                self.combobox_release_upgrades.set_active(i)
-                self.combobox_release_upgrades.handler_unblock(combo_handler)
+
+        def handler():
+            combobox.set_sensitive(True)
+
+        def on_error(e):
+            combobox.set_sensitive(True)
+
+            if isinstance(e, dbus.DBusException):
+                if maybe_log_authentication_canceled_error(e):
+                    self.init_release_upgrades()
+                    return
+            raise e
+
+        combobox.set_sensitive(False)
+        self.backend.SetReleaseUpgradesPolicy(i, reply_handler=handler,
+                                              error_handler=on_error)
 
     def on_combobox_updates_subscription_changed(self, combobox):
         index = combobox.get_active()
@@ -723,14 +773,33 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             except ValueError:
                 pass
 
+        combobox.set_sensitive(False)
+
         def _toggle_child_source(pocket, enabled):
+            loop = GLib.MainLoop()
+            exception = None
+
+            def on_error(e):
+                nonlocal exception
+                exception = e
+                loop.quit()
+
             source = "{}-{}".format(self.distro.codename, pocket)
             if enabled:
                 logging.info("enabling {}".format(source))
-                self.backend.EnableChildSource(source)
+                self.backend.EnableChildSource(source,
+                                               reply_handler=loop.quit,
+                                               error_handler=on_error)
             else:
                 logging.info("disabling {}".format(source))
-                self.backend.DisableChildSource(source)
+                self.backend.DisableChildSource(source,
+                                                reply_handler=loop.quit,
+                                                error_handler=on_error)
+
+            combobox.popdown()
+            loop.run()
+            if exception:
+                raise exception
 
         try:
             if index == UPDATES_ALL:
@@ -752,7 +821,10 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                 if pocket not in (POCKET_SECURITY, POCKET_UPDATES, POCKET_BACKPORTS):
                     _toggle_child_source(pocket, False)
         except dbus.DBusException as e:
-            maybe_log_authentication_canceled_error(e)
+            if not maybe_log_authentication_canceled_error(e):
+                raise e
+        finally:
+            combobox.set_sensitive(True)
 
     def on_combobox_server_changed(self, combobox):
         """
@@ -769,18 +841,23 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             if name == _("Other..."):
                 dialog = DialogMirror(self.window_main,
                                       self.datadir,
-                                      self.distro,
-                                      self.custom_mirrors)
-                res = dialog.run()
-                if res != None:
-                    self.backend.ChangeMainDownloadServer(res)
-                else:
-                    combobox.set_active(self.active_server)
+                                      self.distro)
+                def _on_mirror_selected(url):
+                    try:
+                        if url:
+                            self.backend.ChangeMainDownloadServer(url)
+                        else:
+                            combobox.set_active(self.active_server)
+                    except dbus.DBusException as e:
+                        if not maybe_log_authentication_canceled_error(e):
+                            raise e
+                dialog.present(_on_mirror_selected)
             elif uri != None and len(self.distro.used_servers) > 0:
                 self.active_server = combobox.get_active()
                 self.backend.ChangeMainDownloadServer(uri)
         except dbus.DBusException as e:
-            maybe_log_authentication_canceled_error(e)
+            if not maybe_log_authentication_canceled_error(e):
+                raise e
 
         # mvo: is this still needed?
         #else:
@@ -797,7 +874,8 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             else:
                 self.backend.DisableComponent(comp)
         except dbus.DBusException as e:
-            maybe_log_authentication_canceled_error(e)
+            if not maybe_log_authentication_canceled_error(e):
+                raise e
 
     def on_checkbutton_child_toggled(self, checkbutton, template):
         """
@@ -809,7 +887,8 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             else:
                 self.backend.EnableChildSource(template.name)
         except dbus.DBusException as e:
-            maybe_log_authentication_canceled_error(e)
+            if not maybe_log_authentication_canceled_error(e):
+                raise e
 
     def on_checkbutton_source_code_toggled(self, checkbutton):
         """ Disable or enable the source code for all sources """
@@ -819,7 +898,8 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             else:
                 self.backend.DisableSourceCodeSources()
         except dbus.DBusException as e:
-            maybe_log_authentication_canceled_error(e)
+            if not maybe_log_authentication_canceled_error(e):
+                raise e
 
     def on_checkbutton_popcon_toggled(self, widget):
         """ The user clicked on the popcon paritipcation button """
@@ -840,22 +920,16 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                                           self.get_comparable,
                                           self.datadir,
                                           file)
-            (res, new_sources) = dialog.run()
-            if res == RESPONSE_REPLACE:
-                self.sourceslist.list = []
+            def _on_add_sources_response(res, new_sources):
+                if res == RESPONSE_REPLACE:
+                    self.sourceslist.list = []
                 if res in (RESPONSE_ADD, RESPONSE_REPLACE):
                     for source in new_sources:
                         self.backend.AddSourceFromLine(str(source))
+            dialog.present(_on_add_sources_response)
         except dbus.DBusException as e:
-            maybe_log_authentication_canceled_error(e)
-
-    def on_sources_drag_data_received(self, widget, context, x, y,
-                                      selection, target_type, timestamp):
-        """Extract the dropped file pathes and open the first file, only"""
-        uri = selection.data.strip()
-        uri_splitted = uri.split()
-        if len(uri_splitted) > 0:
-            self.open_file(uri_splitted[0])
+            if not maybe_log_authentication_canceled_error(e):
+                raise e
 
     def hide(self):
         self.window_main.hide()
@@ -892,6 +966,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         cell_toggle = Gtk.CellRendererToggle()
         cell_toggle.set_property("xpad", 2)
         cell_toggle.set_property("ypad", 2)
+        self.maybe_disconnect_handler(cell_toggle)
         self.handlers[cell_toggle] = cell_toggle.connect('toggled',
                                                   self.on_isv_source_toggled)
         col_active = Gtk.TreeViewColumn(_("Active"), cell_toggle,
@@ -908,6 +983,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         cell_toggle = Gtk.CellRendererToggle()
         cell_toggle.set_property("xpad", 2)
         cell_toggle.set_property("ypad", 2)
+        self.maybe_disconnect_handler(cell_toggle)
         self.handlers[cell_toggle] = cell_toggle.connect('toggled',
                                                   self.on_isv_source_toggled,
                                                   self.source_store)
@@ -919,16 +995,6 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         # typeahead support for the sources list using case insensitive matching
         self.treeview_sources.set_search_column(COLUMN_DESC)
         self.treeview_sources.set_search_equal_func(self.source_search_function, None)
-        # drag and drop support for sources.list
-        try:
-            Gtk.drag_dest_set(self.treeview_sources, Gtk.DestDefaults.ALL,
-                              [Gtk.TargetEntry.new('text/uri-list', 0, 0)],
-                              Gdk.DragAction.COPY)
-            self.treeview_sources.connect("drag_data_received",
-                                          self.on_sources_drag_data_received)
-        except AttributeError:
-            # does not work with Gtk2/GI
-            pass
 
     def on_isv_source_activate(self, treeview, path, column):
         """Open the edit dialog if a channel was double clicked"""
@@ -949,7 +1015,8 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         try:
             self.backend.ToggleSourceUse(str(source_entry))
         except dbus.DBusException as e:
-            maybe_log_authentication_canceled_error(e)
+            if not maybe_log_authentication_canceled_error(e):
+                raise e
 
     def init_keys(self):
         """Setup the user interface parts needed for the key handling"""
@@ -958,49 +1025,6 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         tr = Gtk.CellRendererText()
         keys_col = Gtk.TreeViewColumn(_("Key"), tr, markup=0)
         self.treeview_auth.append_column(keys_col)
-        try:
-            self.treeview_auth.enable_model_drag_dest(
-                [('text/plain', 0, 0)], Gdk.DragAction.COPY)
-            self.treeview_auth.connect("drag_data_received",
-                                       self.on_auth_drag_data_received)
-        except AttributeError:
-            # Does not work with GTK 2/GI
-            pass
-
-        self.treeview_auth.connect("button-press-event",
-                                   self.show_auth_context_menu)
-
-    def show_auth_context_menu(self, widget, event):
-        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 3:
-            menu = Gtk.Menu()
-            item_paste = Gtk.MenuItem(label=_("_Add key from paste data"))
-            item_paste.connect("activate", self.on_auth_add_key_from_paste)
-            menu.append(item_paste)
-            menu.show_all()
-            menu.popup(None, None, None, None, event.button, event.time)
-            return True
-
-    def on_auth_add_key_from_paste(self, widget):
-        keydata = Gtk.Clipboard().wait_for_text()
-        if not keydata:
-            return
-        if not self.add_key_from_data(keydata):
-            error(self.window_main,
-                  _("Error importing key"),
-                  _("The selected data may not be a GPG key file "
-                    "or it might be corrupt."))
-        self.show_keys()
-
-    def on_auth_drag_data_received(self, widget, context, x, y,
-                                   selection, target_type, timestamp):
-        """Extract the dropped key and add it to the keyring"""
-        keydata = selection.get_data().strip()
-        if not self.add_key_from_data(keydata):
-            error(self.window_main,
-                  _("Error importing key"),
-                  _("The selected data may not be a GPG key file "
-                    "or it might be corrupt."))
-        self.show_keys()
 
     def add_key_from_data(self, keydata):
         return self.backend.AddKeyFromData(keydata)
@@ -1010,7 +1034,8 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         try:
             self.backend.Revert()
         except dbus.DBusException as e:
-            maybe_log_authentication_canceled_error(e)
+            if not maybe_log_authentication_canceled_error(e):
+                raise e
         if self.settings:
             self.settings.set_int("regular-auto-launch-interval", self.initial_auto_launch)
         self.show_auto_update_level()
@@ -1081,29 +1106,35 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         i = self.combobox_update_interval.get_active()
         if i != -1:
             value = self.combobox_interval_mapping[i]
-            try:
-                self.backend.SetUpdateInterval(value)
-            except dbus.DBusException as e:
-                if maybe_log_authentication_canceled_error(e):
-                    update_days = self.get_update_interval()
-                    combo_handler = self.handlers[self.combobox_update_interval]
-                    for key in self.combobox_interval_mapping:
-        	            if self.combobox_interval_mapping[key] == update_days:
-	                        self.combobox_update_interval.handler_block(combo_handler)
-	                        self.combobox_update_interval.set_active(key)
-	                        self.combobox_update_interval.handler_unblock(combo_handler)
-	                        break
+
+            def handler():
+                widget.set_sensitive(True)
+
+            def on_error(e):
+                widget.set_sensitive(True)
+
+                if isinstance(e, dbus.DBusException):
+                    if maybe_log_authentication_canceled_error(e):
+                        self.init_auto_update()
+                        return
+                raise e
+
+            widget.set_sensitive(False)
+            self.backend.SetUpdateInterval(value, reply_handler=handler,
+                                           error_handler=on_error)
 
     def on_add_clicked(self, widget):
         """Show a dialog that allows to enter the apt line of a to be used repo"""
-        dialog = DialogAdd(self.window_main, self.sourceslist,
-                           self.datadir, self.distro)
-        line = dialog.run()
-        if line != None:
-            try:
-                self.backend.AddSourceFromLine(line)
-            except dbus.DBusException as e:
-                maybe_log_authentication_canceled_error(e)
+        dialog = DialogAdd(self.window_main, self.sourceslist, self.datadir, self.distro)
+
+        def _on_dialog_response(response_id, raw_apt_line):
+            if raw_apt_line is not None:
+                try:
+                    self.backend.AddSourceFromLine(raw_apt_line)
+                except dbus.DBusException as e:
+                    maybe_log_authentication_canceled_error(e)
+
+        dialog.present(_on_dialog_response)
 
     def on_edit_clicked(self, widget):
         """Show a dialog to edit an ISV source"""
@@ -1119,12 +1150,15 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         else:
             dialog = DialogEdit(self.window_main, self.sourceslist,
                                 old_source_entry, self.datadir)
-        if dialog.run() == Gtk.ResponseType.OK:
-            try:
-                self.backend.ReplaceSourceEntry(str(old_source_entry),
-                                                str(dialog.new_source_entry))
-            except dbus.DBusException as e:
-                maybe_log_authentication_canceled_error(e)
+        def _on_edit_response(response_id):
+            if response_id == Gtk.ResponseType.OK:
+                try:
+                    self.backend.ReplaceSourceEntry(str(old_source_entry),
+                                                    str(dialog.new_source_entry))
+                except dbus.DBusException as e:
+                    maybe_log_authentication_canceled_error(e)
+
+        dialog.present(_on_edit_response)
 
     # FIXME:outstanding from merge
     def on_isv_source_activated(self, treeview, path, column):
@@ -1164,19 +1198,26 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
     def add_key_clicked(self, widget):
         """Provide a file chooser that allows to add the gnupg of a trusted
            software vendor"""
-        chooser = Gtk.FileChooserDialog(title=_("Import key"),
-                                        parent=self.window_main,
-                                        buttons=(Gtk.STOCK_CANCEL,
-                                                 Gtk.ResponseType.REJECT,
-                                                 Gtk.STOCK_OK, Gtk.ResponseType.ACCEPT))
+        file_dialog = Gtk.FileDialog(title=_("Import key"))
+
         if "SUDO_USER" in os.environ:
             home = os.path.expanduser("~%s" % os.environ["SUDO_USER"])
-            chooser.set_current_folder(home)
-        res = chooser.run()
-        chooser.hide()
-        if res == Gtk.ResponseType.ACCEPT:
+            file_dialog.set_initial_folder(Gio.File.new_for_path(home))
+
+        def _on_open_finished(dialog, result):
             try:
-                if not self.backend.AddKey(chooser.get_filename()):
+                selected_file = dialog.open_finish(result)
+            except GLib.Error:
+                # Likely the user cancelled us
+                return
+            if not selected_file:
+                return
+            file_path = selected_file.get_path()
+            if not file_path:
+                return
+            try:
+                # FIXME: The backend does not validate that the key contains valid key data...
+                if not self.backend.AddKey(file_path):
                     error(self.window_main,
                           _("Error importing selected file"),
                           _("The selected file may not be a GPG key file "
@@ -1184,6 +1225,8 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             except dbus.DBusException as e:
                 maybe_log_authentication_canceled_error(e)
 
+        file_dialog.open(self.window_main, None, _on_open_finished)
+
     def remove_key_clicked(self, widget):
         """Remove a trusted software vendor key"""
         selection = self.treeview_auth.get_selection()
@@ -1198,23 +1241,27 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                       _("The key you selected could not be removed. "
                         "Please report this as a bug."))
         except dbus.DBusException as e:
-            maybe_log_authentication_canceled_error(e)
+            if not maybe_log_authentication_canceled_error(e):
+                raise e
 
-    def on_delete_event(self, widget, args):
+    def on_close_request(self, widget):
         """Close the window if requested"""
         self.on_close_button(widget)
-        return False
+        # Stop default handler so window isn't destroyed immediately
+        return True
 
     def on_close_button(self, widget):
         """Show a dialog that a reload of the channel information is required
            only if there is no parent defined"""
+        def _on_done():
+            self.quit()
+
         if (self.modified_sourceslist == True and
                 self.options.no_update == False):
-            d = DialogCacheOutdated(self.window_main,
-                                    self.datadir)
-            d.run()
-
-        self.quit()
+            d = DialogCacheOutdated(self.window_main, _on_done, self.datadir)
+            d.present()
+        else:
+            self.quit()
 
     def on_driver_changes_progress(self, progress, ptype, data=None):
         #print(progress)
@@ -1360,13 +1407,16 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         self.button_driver_apply.set_visible(True)
         self.button_driver_cancel.set_visible(False)
         self.button_driver_restart.set_visible(False)
-        self.box_driver_action.pack_end(self.button_driver_apply, False, False, 0)
-        self.box_driver_action.pack_end(self.button_driver_revert, False, False, 0)
-        self.box_driver_action.pack_end(self.button_driver_restart, False, False, 0)
-        self.box_driver_action.pack_end(self.button_driver_cancel, False, False, 0)
+        self.box_driver_action.append(self.button_driver_apply)
+        self.box_driver_action.append(self.button_driver_revert)
+        self.box_driver_action.append(self.button_driver_restart)
+        self.box_driver_action.append(self.button_driver_cancel)
 
         self.progress_bar = Gtk.ProgressBar()
-        self.box_driver_action.pack_end(self.progress_bar, False, False, 0)
+        self.box_driver_action.insert_child_after(self.progress_bar, self.label_driver_action)
+        self.progress_bar.set_valign(Gtk.Align.CENTER)
+        self.progress_bar.set_halign(Gtk.Align.START)
+        self.progress_bar.set_hexpand(True)
         self.progress_bar.set_visible(False)
 
         self.devices = {}
@@ -1613,31 +1663,31 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             driver_status = Gtk.Image()
             driver_status.set_valign(Gtk.Align.START)
             driver_status.set_halign(Gtk.Align.CENTER)
-            driver_status.set_from_icon_name(icon, Gtk.IconSize.MENU)
+            driver_status.set_from_icon_name(icon)
             device_box = Gtk.Box(spacing=6, orientation=Gtk.Orientation.HORIZONTAL)
-            device_box.pack_start(driver_status, False, False, 6)
+            device_box.append(driver_status)
             device_detail = Gtk.Box(spacing=6, orientation=Gtk.Orientation.VERTICAL)
-            device_box.pack_start(device_detail, True, True, 0)
+            device_box.append(device_detail)
 
             widget = Gtk.Label.new("{}: {}".format(self.devices[device].get('vendor', _('Unknown')), self.devices[device].get('model', _('Unknown'))))
             widget.set_halign(Gtk.Align.START)
-            device_detail.pack_start(widget, True, False, 0)
+            device_detail.append(widget);
             widget = Gtk.Label.new("<small>{}</small>".format(overall_status))
             widget.set_halign(Gtk.Align.START)
             widget.set_use_markup(True)
-            device_detail.pack_start(widget, True, False, 0)
+            device_detail.append(widget)
             self.dynamic_device_status[device] = (driver_status, widget)
 
             option_group = None
             # define the order of introspection
             for section in ('recommended', 'alternative', 'manually_installed', 'no_driver'):
                 for driver in drivers[section]:
-                    radio_button = Gtk.RadioButton.new_with_label(None, drivers[section][driver]['description'])
+                    radio_button = Gtk.CheckButton.new_with_label(drivers[section][driver]['description'])
                     if option_group:
-                        radio_button.join_group(option_group)
+                        radio_button.set_group(option_group)
                     else:
                         option_group = radio_button
-                    device_detail.pack_start(radio_button, True, False, 0)
+                    device_detail.append(radio_button)
                     radio_button.set_active(drivers[section][driver]['selected'])
 
                     if section == 'no_driver':
@@ -1649,10 +1699,10 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                     if drivers['manually_installed'] and section != 'manually_installed':
                         radio_button.set_sensitive(False)
 
-            self.box_driver_detail.pack_start(device_box, False, False, 6)
+            self.box_driver_detail.append(device_box)
 
         self.ui_building = False
-        self.box_driver_detail.show_all()
+        self.box_driver_detail.show()
         self.set_driver_action_status()
 
     def update_label_and_icons_from_status(self):
@@ -1662,7 +1712,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             (overall_status, icon, drivers) = self.gather_device_data(self.devices[device])
             (driver_status, widget) = self.dynamic_device_status[device]
 
-            driver_status.set_from_icon_name(icon, Gtk.IconSize.MENU)
+            driver_status.set_from_icon_name(icon)
             widget.set_label("<small>{}</small>".format(overall_status))
 
     def set_driver_action_status(self):
@@ -1670,7 +1720,6 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         if (os.path.exists('/var/run/reboot-required')):
             self.label_driver_action.set_label(_("You need to restart the computer to complete the driver changes."))
             self.button_driver_restart.set_visible(True)
-            self.window_main.set_urgency_hint(True)
             return
 
         self.nonfree_drivers = 0
