@@ -17,12 +17,13 @@
 # 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 
 import aptsources.distro
+import ctypes
 from datetime import datetime
 import distro_info
 from functools import wraps
 import gi
 gi.require_version("Gtk", "4.0")
-from gi.repository import Gio, Gtk, GObject
+from gi.repository import Gio, Gtk, Gdk, GdkX11, GObject
 import json
 import os
 import subprocess
@@ -175,3 +176,61 @@ def retry(exceptions, tries=10, delay=0.1, backoff=2):
         return f_retry  # true decorator
 
     return deco_retry
+
+
+def maybe_hint_dialog_window(window):
+    """
+    Workaround for GTK4 removal of set_type_hint.
+    Manually sets _NET_WM_WINDOW_TYPE_DIALOG on X11.
+    See: https://discourse.gnome.org/t/replacing-gtk-window-set-type-hint-in-gtk4
+    """
+    display = Gdk.Display.get_default()
+    # Return early if we're not running on X11
+    if not isinstance(display, GdkX11.X11Display):
+        return
+
+    def _apply_x11_hint(win):
+        native = win.get_native()
+        surface = native.get_surface()
+        xid = surface.get_xid()
+
+        # Load X11 (Allow this to raise OSError if missing, per review)
+        libx11 = ctypes.CDLL("libX11.so.6")
+
+        XA_ATOM = 4
+        PropModeReplace = 0
+        libx11.XOpenDisplay.restype = ctypes.c_void_p
+        libx11.XInternAtom.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_int]
+        libx11.XInternAtom.restype = ctypes.c_ulong
+
+        # Open a new connection to avoid any pointer compatibility
+        # issues between gi and ctypes.
+        display_ptr = libx11.XOpenDisplay(None)
+        if not display_ptr:
+            return
+
+        try:
+            atom_type = libx11.XInternAtom(display_ptr, b"_NET_WM_WINDOW_TYPE", 0)
+            atom_dialog = libx11.XInternAtom(display_ptr, b"_NET_WM_WINDOW_TYPE_DIALOG", 0)
+            data = (ctypes.c_ulong * 1)(atom_dialog)
+
+            # XChangeProperty(display, w, property, type, format, mode, data, nelements)
+            libx11.XChangeProperty(
+                display_ptr,
+                xid,
+                atom_type,
+                XA_ATOM,
+                32,
+                PropModeReplace,
+                ctypes.cast(data, ctypes.POINTER(ctypes.c_ubyte)),
+                1
+            )
+            libx11.XFlush(display_ptr)
+        finally:
+            libx11.XCloseDisplay(display_ptr)
+
+    # Apply hints immediately if already realised, otherwise wait for signal
+    if window.get_realized():
+        _apply_x11_hint(window)
+    else:
+        window.connect('realize', _apply_x11_hint)
