@@ -2,18 +2,20 @@
 
 import apt
 
+from functools import partial
 import unittest
 import sys
 import os
 
 from aptsources.distro import get_distro
-from softwareproperties.extendedsourceslist import SourceEntry
+from aptsources.sourceslist import (SourceEntry, Deb822SourceEntry)
 from contextlib import contextmanager
 from http.client import HTTPException
 from launchpadlib.launchpad import Launchpad
-from mock import (patch, Mock)
+from lazr.restfulclient.errors import Unauthorized
 from urllib.request import urlopen
 from urllib.error import URLError
+from unittest.mock import (patch, Mock)
 
 sys.path.insert(0, "..")
 
@@ -21,7 +23,7 @@ from softwareproperties.sourceslist import SourcesListShortcutHandler
 from softwareproperties.uri import URIShortcutHandler
 from softwareproperties.cloudarchive import CloudArchiveShortcutHandler
 from softwareproperties.ppa import PPAShortcutHandler
-from softwareproperties.shortcuthandler import InvalidShortcutException
+from softwareproperties.shortcuthandler import InvalidShortcutException, ShortcutException
 from softwareproperties.shortcuts import shortcut_handler
 
 
@@ -29,32 +31,38 @@ DISTRO = get_distro()
 CODENAME = DISTRO.codename
 
 # These must match the ppa used in the VALID_PPAS
-PPA_LINE = f"deb http://ppa.launchpad.net/ddstreet/ppa/ubuntu/ {CODENAME} main"
+PPA_LINE = f"""Types: deb
+URIs: https://ppa.launchpadcontent.net/ddstreet/ppa/ubuntu/
+Suites: {CODENAME}
+Components: main
+"""
 PPA_FILEBASE = "ddstreet-ubuntu-ppa"
-PPA_SOURCEFILE = f"{PPA_FILEBASE}-{CODENAME}.list"
-PPA_TRUSTEDFILE = f"{PPA_FILEBASE}.gpg"
+PPA_SOURCEFILE = f"{PPA_FILEBASE}-{CODENAME}.sources"
 PPA_NETRCFILE = f"{PPA_FILEBASE}.conf"
 
 PRIVATE_PPA_PASSWORD = "thisisnotarealpassword"
-PRIVATE_PPA_LINE = f"deb https://private-ppa.launchpad.net/ddstreet/ppa/ubuntu/ {CODENAME} main"
-PRIVATE_PPA_NETRCCONTENT = f"machine private-ppa.launchpad.net/ddstreet/ppa/ubuntu/ login ddstreet password {PRIVATE_PPA_PASSWORD}"
-PRIVATE_PPA_SUBSCRIPTION_URLS = [f"https://ddstreet:{PRIVATE_PPA_PASSWORD}@private-ppa.launchpad.net/ddstreet/ppa/ubuntu/"]
+PRIVATE_PPA_LINE = f"""Types: deb
+URIs: https://private-ppa.launchpadcontent.net/ddstreet/ppa/ubuntu/
+Suites: {CODENAME}
+Components: main
+"""
+PRIVATE_PPA_NETRCCONTENT = f"machine private-ppa.launchpadcontent.net/ddstreet/ppa/ubuntu/ login ddstreet password {PRIVATE_PPA_PASSWORD}"
+PRIVATE_PPA_SUBSCRIPTION_URL = f"https://ddstreet:{PRIVATE_PPA_PASSWORD}@private-ppa.launchpadcontent.net/ddstreet/ppa/ubuntu/"
 
 # These must match the uca used in VALID_UCAS
 UCA_CANAME = "train"
 UCA_ARCHIVE = "http://ubuntu-cloud.archive.canonical.com/ubuntu"
-UCA_LINE = f"deb {UCA_ARCHIVE} bionic-updates/{UCA_CANAME} main"
-UCA_LINE_PROPOSED = f"deb {UCA_ARCHIVE} bionic-proposed/{UCA_CANAME} main"
+UCA_LINE = f"deb {UCA_ARCHIVE} {CODENAME}-updates/{UCA_CANAME} main"
+UCA_LINE_PROPOSED = f"deb {UCA_ARCHIVE} {CODENAME}-proposed/{UCA_CANAME} main"
 UCA_FILEBASE = f"cloudarchive-{UCA_CANAME}"
 UCA_SOURCEFILE = f"{UCA_FILEBASE}.list"
-CA_ALLOW_CODENAME = "bionic"
 
 # This must match the VALID_URIS
 URI = "http://fake.mirror.private.com/ubuntu"
 URI_FILEBASE = "archive_uri-http_fake_mirror_private_com_ubuntu"
 URI_SOURCEFILE = f"{URI_FILEBASE}-{CODENAME}.list"
 
-VALID_LINES = [f"deb {URI} bionic main"]
+VALID_LINES = [f"deb {URI} {CODENAME} main"]
 VALID_URIS = [URI]
 VALID_PPAS = ["ppa:ddstreet", "ppa:~ddstreet", "ppa:ddstreet/ppa", "ppa:~ddstreet/ppa", "ppa:ddstreet/ubuntu/ppa", "ppa:~ddstreet/ubuntu/ppa"]
 VALID_UCAS = [f"cloud-archive:{UCA_CANAME}", f"cloud-archive:{UCA_CANAME}-updates", f"cloud-archive:{UCA_CANAME}-proposed", f"uca:{UCA_CANAME}", f"uca:{UCA_CANAME}-updates", f"uca:{UCA_CANAME}-proposed"]
@@ -75,12 +83,13 @@ def has_network():
     return True
 
 def mock_login_with(*args, **kwargs):
+    has_subscription = kwargs.pop("has_subscription", True)
+
     _lp = Launchpad.login_anonymously(*args, **kwargs)
     lp = Mock(wraps=_lp)
 
     lp.me = Mock()
     lp.me.name = 'ddstreet'
-    lp.me.getArchiveSubscriptionURLs = lambda: PRIVATE_PPA_SUBSCRIPTION_URLS
 
     def mock_getPPAByName(_team, name):
         _ppa = _team.getPPAByName(name=name)
@@ -93,6 +102,15 @@ def mock_login_with(*args, **kwargs):
         _team = _lp.people(teamname)
         team = Mock(wraps=_team)
         team.getPPAByName = lambda name: mock_getPPAByName(_team, name)
+        team.getArchiveSubscriptionURL = lambda archive: PRIVATE_PPA_SUBSCRIPTION_URL
+        if has_subscription:
+            team.getArchiveSubscriptionURL = lambda archive: PRIVATE_PPA_SUBSCRIPTION_URL
+        else:
+            def raise_unauthorized(archive):
+                raise Unauthorized(Mock(), Mock())
+
+            team.getArchiveSubscriptionURL = raise_unauthorized
+
         return team
 
     lp.people = mock_people
@@ -139,12 +157,22 @@ class ShortcutsTestcase(unittest.TestCase):
             self.assertEqual(shortcut.sourceparts_filename, sourcefile)
             self.assertEqual(shortcut.sourceparts_file, os.path.join(sourceparts, sourcefile))
 
-        binentry = SourceEntry(line)
-        binentry.type = DISTRO.binary_type
+        if shortcut.deb822:
+            binentry = Deb822SourceEntry(line, '')
+            binentry.types = [DISTRO.binary_type]
+        else:
+            binentry = SourceEntry(line)
+            binentry.type = DISTRO.binary_type
+
         self.assertEqual(shortcut.SourceEntry(shortcut.binary_type), binentry)
 
-        srcentry = SourceEntry(line)
-        srcentry.type = DISTRO.source_type
+        if shortcut.deb822:
+            srcentry = Deb822SourceEntry(line, '')
+            srcentry.types = [DISTRO.source_type]
+        else:
+            srcentry = SourceEntry(line)
+            srcentry.type = DISTRO.source_type
+
         srcentry.set_enabled(self.enable_source)
         self.assertEqual(shortcut.SourceEntry(shortcut.source_type), srcentry)
 
@@ -183,7 +211,6 @@ class ShortcutsTestcase(unittest.TestCase):
             for shortcut in self.create_handlers(ppa, PPAShortcutHandler):
                 self.check_shortcut(shortcut, PPA_LINE,
                                     sourcefile=PPA_SOURCEFILE,
-                                    trustedfile=PPA_TRUSTEDFILE,
                                     netrcfile=PPA_NETRCFILE,
                                     trustedcontent=True)
 
@@ -196,11 +223,19 @@ class ShortcutsTestcase(unittest.TestCase):
                 for shortcut in self.create_handlers(ppa, PPAShortcutHandler, login=True):
                     self.check_shortcut(shortcut, PRIVATE_PPA_LINE,
                                         sourcefile=PPA_SOURCEFILE,
-                                        trustedfile=PPA_TRUSTEDFILE,
                                         netrcfile=PPA_NETRCFILE,
                                         trustedcontent=True,
                                         netrccontent=PRIVATE_PPA_NETRCCONTENT)
 
+    @unittest.skipUnless(has_network(), "requires network")
+    def test_shortcut_private_ppa_no_subscription(self):
+        # this is the same tests as the public ppa, but login=True will use the mocked lp instance
+        # this *does not* actually test/verify this works with a real private ppa; that must be done manually
+        with patch('launchpadlib.launchpad.Launchpad.login_with', new=partial(mock_login_with, has_subscription=False)):
+            for ppa in VALID_PPAS:
+                with self.assertRaises(ShortcutException):
+                    self.create_handler(ppa, PPAShortcutHandler, login=True)
+
     @contextmanager
     def ca_allow_codename(self, codename):
         key = "CA_ALLOW_CODENAME"
@@ -217,7 +252,7 @@ class ShortcutsTestcase(unittest.TestCase):
     def test_shortcut_cloudarchive(self):
         for uca in VALID_UCAS:
             line = UCA_LINE_PROPOSED if 'proposed' in uca else UCA_LINE
-            with self.ca_allow_codename(CA_ALLOW_CODENAME):
+            with self.ca_allow_codename(CODENAME):
                 for shortcut in self.create_handlers(uca, CloudArchiveShortcutHandler):
                     self.check_shortcut(shortcut, line, sourcefile=UCA_SOURCEFILE)
 
