@@ -5,13 +5,14 @@ import apt_pkg
 import aptsources
 
 import logging
-from mock import patch
 import os
 import shutil
 import sys
 import tempfile
 import unittest
 
+from unittest.mock import patch
+
 sys.path.insert(0, "../")
 from softwareproperties.SoftwareProperties import (
     SoftwareProperties)
